/*********************************************************************
    This file is part of Siworin. Copyright © 2021 by Ruud Harmsen.

    Siworin is a simple word indexer and local search engine. See
    https://rudhar.com/sfreview/siworin/#siw04 and
    https://rudhar.com/sfreview/siworin/toolsrc/siworin .

    Siworin is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    Siworin is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Siworin.  If not, see <https://www.gnu.org/licenses/>.
  ********************************************************************/

/* Siworin - simple word indexer
   Written by Ruud Harmsen, August 2021
   CGI-interface for searching the word index built up from HTML files.

   16 September 2022: Measures against Cross-Site Scripting (XSS).
	See function against_XSS, and see:
	https://www.openbugbounty.org/
	https://cwe.mitre.org/data/definitions/79.html
	https://cheatsheetseries.owasp.org/cheatsheets/Cross_Site_Scripting_Prevention_Cheat_Sheet.html

   15 January 2023: Offending character are removed only for display,
	but they are still used for searching.
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <wctype.h>
#include <locale.h>
#include <unistd.h>
#include "cgitools.h"
#include "siworin.h"
#include "siworin-files.h"
#include "siworin-config.h"
#include "siworin-srchind.h"

static struct siworin_formdata_s formdata;

static void showform (char *script, struct siworin_formdata_s *p_formdata);
static void htmlhead (void);
static void htmlfoot (void);
static void get_invocation_parms (struct siworin_formdata_s *p_formdata);
static void getparms (int len, struct siworin_formdata_s *p_formdata);
static int  searchandshow (struct siworin_formdata_s *p_formdata);
static void against_XSS (char *input);

int main (void)
{
   int len;
   char *scriptname = "../cgi-bin/s.cgi";

   /* Stderr, e.g. for perror in the macro SIWORIN_FAIL
      (#defined in siworin.h) isn't visible anywhere in
      a CGI-program. Therefore I redirect it to stdout,
      so any error reports appear on the screen. */
   /* close(fileno(stderr)); dup(fileno(stdout)); */
   /* Atomic equivalent of the previous. */
   dup2(fileno(stdout), fileno(stderr));

   setlocale(LC_ALL, "en_US.UTF-8");

   get_invocation_parms(&formdata);

   if (formdata.searchstring[0])
   {
      /* Search pattern set from URL invocation. Immediately
         show search results based on that. */
      htmlhead();
      showform(scriptname, &formdata);
      searchandshow(&formdata);
      htmlfoot();
   }
   else if ((len = userinput()) == 0)
   {
      /* First call, with empty search pattern, still to be set
         by the user. */
      htmlhead();
      /* Default value is: search combining search words */
      formdata.combin = 1;

      showform(scriptname, &formdata);
      htmlfoot();
   }
   else
   {
      /* User entered a searchstring, and possibly some parameters,
         and wants to see results. */
      getparms(len, &formdata);
      htmlhead();
      showform(scriptname, &formdata);
      searchandshow(&formdata);
      htmlfoot();
   }
   siworin_free_index(&formdata);

   return 0;
}

static void htmlhead (void)
{
   FILE *fp;
   char *fn;
   int c;

   if (!(fn = siworin_config_value(SIWORIN_HEAD_HTML_FILE)))
      SIWORIN_FAIL("get head html file");
   if (!(fp = fopen(fn, "r")))
      SIWORIN_FAIL("open head html file");

   while ((c = getc(fp)) != EOF)
      fputc(c, stdout);
}

static void showform (char *script, struct siworin_formdata_s *p_formdata)
{
   char explicit_config[128];
   char *confkey = "?config=";

   explicit_config[0] = '\0';
   if (siworin_get_explicit_config_path())
   {
      strcpy(explicit_config, confkey);
      strncat(explicit_config, siworin_get_explicit_config_path(),
         sizeof explicit_config - strlen(confkey) - 1);
   }

   lowercase_searchstring(p_formdata->searchstring);
   strcpy(p_formdata->dsplaystring, p_formdata->searchstring);
   against_XSS(p_formdata->dsplaystring);

   siworin_index_srchwrds(p_formdata);

   fprintf(stdout,
      "<form action=\"%s%s\" method=post autocomplete=on>\n"
      "<br>\n"
      "<input type=Checkbox name=regexp %s>Regular expression\n"
      "<input type=Checkbox name=combin %s>Combined<br>\n"
      "<input type=Text name=find  size=55 maxlength=%d value=\"%s\" autofocus>\n"
      "<input type=Submit value=\"Go!\">\n"
      "</p>\n</form>\n\n",
      script, explicit_config,
      p_formdata->regexp ? "checked" : "",
      p_formdata->combin ? "checked" : "",
      (int)(sizeof p_formdata->dsplaystring) - 1,
      p_formdata->dsplaystring);

   fflush(stdout);
}

static void htmlfoot (void)
{
   fprintf(stdout, "</body></html>\n");
   fflush(stdout);
}

static void get_invocation_parms (struct siworin_formdata_s *p_formdata)
{
   char *query;
   char **Vars = NULL, **V;

   memset(p_formdata, '\0', sizeof *p_formdata);

   query = getenv("QUERY_STRING");
   if (query)
   {
      Vars = CgiConvVars(query, NULL);

      for (V = Vars; V[0]; V += 2)
      {
         if ((!strcmp(V[0], "find") || !strcmp(V[0], "q")) && strlen(V[1]))
         {
            strncpy(   p_formdata->searchstring, V[1],
               (sizeof p_formdata->searchstring) - 1);
         }
         else if (strcmp(V[0], "regexp") == 0 && strcmp(V[1], "1") == 0)
         {
            p_formdata->regexp = 1;
         }
         else if (strcmp(V[0], "combin") == 0 && strcmp(V[1], "1") == 0)
         {
            p_formdata->combin = 1;
         }
         else if (strcmp(V[0], "config") == 0)
         {
            siworin_set_explicit_config_path(V[1]);
         }
      }
   }
}

static void getparms (int len, struct siworin_formdata_s *p_formdata)
{
   char **Vars = NULL, **V;
   char *buf = calloc(len + 1, 1);

   if (!buf || fread(buf, len, 1, stdin) != 1)
   {
      free(buf);
      return;
   }

   Vars = CgiConvVars(buf, NULL);

   for (V = Vars; V[0]; V += 2)
   {
      if (strcmp(V[0], "regexp") == 0 && strlen(V[1]))
         p_formdata->regexp = ((strcmp(V[1], "on") == 0) ? 1 : 0);
      else
      if (strcmp(V[0], "combin") == 0 && strlen(V[1]))
         p_formdata->combin = ((strcmp(V[1], "on") == 0) ? 1 : 0);
      else
      if (strcmp(V[0], "find") == 0 && strlen(V[1]))
         strncpy(p_formdata->searchstring, V[1], sizeof p_formdata->searchstring - 1);
   }

   free(buf);
   if (Vars)
      free(Vars);
}

static int searchandshow (struct siworin_formdata_s *p_formdata)
{
   if (siworin_fileopen_paths(
       siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX),
       "r", "r"))
      SIWORIN_FAIL("fopens");
   if (siworin_fileopen_words(
       siworin_config_value(SIWORIN_CONFIG_INDEX_FILES_PREFIX),
       "r", "r"))
      SIWORIN_FAIL("fopens");

   siworin_handle_srchwrds(p_formdata);

   siworin_fileopen_close();

   return 0;
}

/* Wipe out any < and > etc., in input entered by the user,
   or introduced via the URI, to avoid maliciously input
   HTML code, and possible scripting languages within that.
   The cleaned string, with offending characters replaced
   by dots (for improved chances of finding thing with
   regular regular expressions) are used for display.

   Hopefully this will make XSS (cross-site scripting)
   impossible.
 */
static void against_XSS (char *input)
{
   char *to_erase = "<>=\"\'&";
   char *p, *q;

   for (p = input; *p; p++)
      for (q = to_erase; *q; q++)
      /* Turn them into a dot, not a space, so they
         may still be findable using regular expressions. */
         if (*p == *q) *p = '.';
}
